/**
 * @file dlist.h
 * Doubly-linked list
 * @author Marko Mäkelä (marko.makela@iki.fi)
 */

/* Copyright © 2007 Marko Mäkelä.

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

/** Item of a doubly-linked list. */
struct dlist_item
{
  /** Previous item */
  struct dlist_item* prev;
  /** Next item */
  struct dlist_item* next;
};

/** Base node of a doubly-linked list. */
struct dlist_base
{
  /** Number of entries */
  int size;
  /** Pointer to first item */
  struct dlist_item* head;
  /** Pointer to last item */
  struct dlist_item* tail;
};

/** Initializer of dlist_base */
#define DLIST_BASE_INITIALIZER { 0, 0, 0 }

/** Add an item as the last element of a list.
 * @param list	base node
 * @param item	item to be added
 */
#define dlist_add_last(list, item)			\
do {							\
  (list)->size++;					\
  (item)->next = 0;					\
  (item)->prev = (list)->tail;				\
  if ((item)->prev) (item)->prev->next = (item);	\
  (list)->tail = (item);				\
  if (!(list)->head) (list)->head = (item);		\
} while (0)

/** Remove an item from the head of a list.
 * @param list	base node
 * @param item	(out) removed item
 */
#define dlist_remove_first(list, item)			\
do {							\
  (item) = (list)->head;				\
  if (item) {						\
    (list)->head = (item)->next;			\
    if ((list)->tail == (item)) (list)->tail = 0;	\
    (list)->size--;					\
  }							\
} while (0)
